@@ -632,34 +632,35 @@ sub setup_tests {
 sub have_encoder_and_decoder {
     my ($min_v)= @_;
     # $Class is the already-loaded class, so the one we're testing
-    my $need = $Class =~ /Encoder/ ? "Decoder" : "Encoder";
-    my $need_class = "Sereal::$need";
+    my @need = $Class =~ /Encoder/ ? ("Decoder") :
+               $Class =~ /Decoder/ ? ("Encoder") :
+                                     ("Encoder", "Decoder");
+    my @need_class = ($Class, map { "Sereal::$_" } @need);
 
-    eval "use $Class; 1"
-    or do {
-        note("Could not locate $Class for testing" . ($@ ? " (Exception: $@)" : ""));
-        return();
-    };
+    foreach my $class (@need_class) {
+        eval "use $class; 1"
+        or do {
+            note("Could not locate $class for testing" . ($@ ? " (Exception: $@)" : ""));
+            return();
+        };
+        my $cmp_v= $class->VERSION;
 
-    eval "use $need_class; 1"
-    or do {
-        note("Could not locate $need_class for testing" . ($@ ? " (Exception: $@)" : ""));
-        return();
-    };
-    my $cmp_v = $need_class->VERSION;
-    if ($min_v and $cmp_v < $min_v) {
-        diag("Could not load correct version of $need_class for testing "
-             ."(got: $cmp_v, needed at least $min_v)");
-        return;
-    }
-    $cmp_v =~ s/_//;
-    $cmp_v = sprintf("%.2f", int($cmp_v*100)/100);
-    my %compat_versions = map {$_ => 1} $Class->_test_compat();
-    if (not defined $cmp_v or not exists $compat_versions{$cmp_v}) {
-        diag("Could not load correct version of $need_class for testing "
-             ."(got: $cmp_v, needed any of ".join(", ", keys %compat_versions).")");
-        return();
+        if ($min_v and $cmp_v < $min_v) {
+            diag("Could not load correct version of $class for testing "
+                 ."(got: $cmp_v, needed at least $min_v)");
+            return;
+        }
+
+        $cmp_v =~ s/_//;
+        $cmp_v = sprintf("%.2f", int($cmp_v*100)/100);
+        my %compat_versions = map {$_ => 1} $Class->_test_compat();
+        if (not defined $cmp_v or not exists $compat_versions{$cmp_v}) {
+            diag("Could not load correct version of $class for testing "
+                 ."(got: $cmp_v, needed any of ".join(", ", keys %compat_versions).")");
+            return();
+        }
     }
+
     return 1;
 }
 
@@ -1054,6 +1055,8 @@ sub run_roundtrip_tests_internal {
     my ($ename, $opt, $encode_decode_callbacks) = @_;
     require Data::Dumper;
 
+    my $failed = 0;
+
     my $decoder = Sereal::Decoder->new($opt);
     my $encoder = Sereal::Encoder->new($opt);
     my %seen_name;
@@ -1061,6 +1064,11 @@ sub run_roundtrip_tests_internal {
     foreach my $rt (@RoundtripTests) {
         my ($name, $data) = @$rt;
 
+        if ($failed > 20) {
+            fail("too many test failures to continue");
+            last;
+        }
+
         TODO:
         foreach my $meth (
               ['object-oriented',
@@ -1089,13 +1097,15 @@ sub run_roundtrip_tests_internal {
             eval {$encoded = $enc->($data); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while encoding: $err");
+                    fail("$name ($ename, $mname, encoding failed)");
+                    $failed++;
                 };
 
             defined($encoded)
                 or do {
                     fail("$name ($ename, $mname, encoded defined)");
                     debug_checks(\$data, \$encoded, undef);
+                    $failed++;
                     next; #test
                 };
 
@@ -1103,13 +1113,16 @@ sub run_roundtrip_tests_internal {
             eval {$decoded = $dec->($encoded); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while decoding: $err");
+                    fail("$name ($ename, $mname, decoding failed)");
+                    $failed++;
+                    next;
                 };
 
             defined($decoded) == defined($data)
                 or do {
                     fail("$name ($ename, $mname, decoded definedness)");
                     debug_checks(\$data, \$encoded, undef);
+                    $failed++;
                     next; #test
                 };
 
@@ -1118,13 +1131,15 @@ sub run_roundtrip_tests_internal {
             eval {$encoded2 = $enc->($decoded); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while encoding the second time: $err");
+                    fail("$name ($ename, $mname, second encoding failed)");
+                    $failed++;
+                    next; #test
                 };
 
             defined $encoded2
                 or do {
                     fail("$name ($ename, $mname, encoded2 defined)");
-                    debug_checks(\$data, \$encoded, \$decoded);
+                    $failed++;
                     next; #test
                 };
 
@@ -1132,13 +1147,15 @@ sub run_roundtrip_tests_internal {
             eval {$decoded2 = $dec->($encoded2); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while decoding the second time: $err");
-                    # hobodecode($encoded2);
+                    fail("$name ($ename, $mname, second decoding failed)");
+                    $failed++;
+                    next; #test
                 };
 
             defined($decoded2) == defined($data)
                 or do {
                     fail("$name ($ename, $mname, decoded2 defined)");
+                    $failed++;
                     next; #test
                 };
 
@@ -1147,13 +1164,15 @@ sub run_roundtrip_tests_internal {
             eval {$encoded3 = $enc->($decoded2); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while encoding the third time: $err");
+                    fail("$name ($ename, $mname, third encoding failed)");
+                    $failed++;
+                    next; #test
                 };
 
             defined $encoded3
                 or do {
                     fail("$name ($ename, $mname, encoded3 defined)");
-                    debug_checks(\$data, \$encoded, \$decoded);
+                    $failed++;
                     next; #test
                 };
 
@@ -1161,31 +1180,38 @@ sub run_roundtrip_tests_internal {
             eval {$decoded3 = $dec->($encoded3); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while decoding the third time: $err");
+                    fail("$name ($ename, $mname, third decoding failed)");
+                    $failed++;
+                    next; #test
                 };
 
             defined($decoded3) == defined($data)
                 or do {
                     fail("$name ($ename, $mname, decoded3 defined)");
+                    $failed++;
                     next; #test
                 };
 
-            deep_cmp($decoded, $data,       "$name ($ename, $mname, decoded vs data)") or next; #test
-            deep_cmp($decoded2, $data,      "$name ($ename, $mname, decoded2 vs data)") or next; #test
-            deep_cmp($decoded2, $decoded,   "$name ($ename, $mname, decoded2 vs decoded)") or next; #test
-
-            deep_cmp($decoded3, $data,      "$name ($ename, $mname, decoded3 vs data)") or next; #test
-            deep_cmp($decoded3, $decoded,   "$name ($ename, $mname, decoded3 vs decoded)") or next; #test
-            deep_cmp($decoded3, $decoded2,  "$name ($ename, $mname, decoded3 vs decoded2)") or next; #test
+            deep_cmp($decoded, $data,       "$name ($ename, $mname, decoded vs data)")
+                or do { $failed++; next }; #test
+            deep_cmp($decoded2, $data,      "$name ($ename, $mname, decoded2 vs data)")
+                or do { $failed++; next }; #test
+            deep_cmp($decoded2, $decoded,   "$name ($ename, $mname, decoded2 vs decoded)")
+                or do { $failed++; next }; #test
+            deep_cmp($decoded3, $data,      "$name ($ename, $mname, decoded3 vs data)")
+                or do { $failed++; next }; #test
+            deep_cmp($decoded3, $decoded,   "$name ($ename, $mname, decoded3 vs decoded)")
+                or do { $failed++; next }; #test
+            deep_cmp($decoded3, $decoded2,  "$name ($ename, $mname, decoded3 vs decoded2)")
+                or do { $failed++; next }; #test
 
             if ( $ename =~ /canon/ ) {
-                deep_cmp($encoded2, $encoded,  "$name ($ename, $mname, encoded2 vs encoded)") or do {
-                    diag Dumper($encoded2);
-                    diag Dumper($encoded);
-                    next; #test
-                };
-                deep_cmp($encoded3, $encoded2, "$name ($ename, $mname, encoded3 vs encoded2)") or next; #test
-                deep_cmp($encoded3, $encoded,  "$name ($ename, $mname, encoded3 vs encoded)") or next; #test
+                deep_cmp($encoded2, $encoded,  "$name ($ename, $mname, encoded2 vs encoded)")
+                    or do { $failed++; next }; #test
+                deep_cmp($encoded3, $encoded2, "$name ($ename, $mname, encoded3 vs encoded2)")
+                    or do { $failed++; next }; #test
+                deep_cmp($encoded3, $encoded,  "$name ($ename, $mname, encoded3 vs encoded)")
+                    or do { $failed++; next }; #test
 
                 if ($ENV{SEREAL_TEST_SAVE_OUTPUT} and $mname eq 'object-oriented') {
                     use File::Path;
@@ -1237,13 +1263,13 @@ sub _write_file {
 # For bootstrapping other language implementations' tests
 our $COMPRESS;
 sub write_test_files {
-    my $dir = shift;
+    my ($dir, $version) = @_;
     require File::Path;
     File::Path::mkpath($dir);
     my $make_data_file_name = sub {File::Spec->catfile($dir, sprintf("test_data_%05u", shift))};
     my $make_name_file_name = sub {File::Spec->catfile($dir, sprintf("test_name_%05u", shift))};
 
-    setup_tests();
+    setup_tests($version);
     foreach my $testno (1..@BasicTests) {
         my $t = $BasicTests[$testno-1];
         my $data = ref($t->[1]) eq 'CODE' ? $t->[1]->() : $t->[1];
